import functools
import os
import posix_ipc
from hm_pyhelper.logger import get_logger

LOGGER = get_logger(__name__)

LOCK_ECC = 'LOCK_ECC'
DEFAULT_TIMEOUT = 2.0  # 2 seconds


class InterprocessLock(object):
    _prefix = "LockSingleton."
    _mode = 0o644

    def __init__(self, name, available_resources=1, reset=False):
        """
        name: uniquely identifies the `LockSingleton` across processes
        available_resources: the count of the available resources
        reset: set True to reset the available_resources

        The available_resources of a `InterprocessLock` get reset on every
        restart of the system or docker container. It's tested in Ubuntu
        20.04 desktop and diagnostics container in a Hotspot.
        Resetting the available_resources by passing the `reset=True` should be
        used with a caution and it can be used in a very specific scenarios
        such as in the development environment. It's designed for facilitating
        the development. It's not recommended to be used in production.
        """
        self._name = self._prefix + name
        # so it doesn't interfere with our semaphore mode
        old_umask = os.umask(0)
        try:
            self._sem = posix_ipc.Semaphore(self._name,
                                            mode=self._mode,
                                            flags=posix_ipc.O_CREAT,
                                            initial_value=available_resources)

            if reset:
                """ Some hack to set the Semaphore's read-only value
https://github.com/rwarren/SystemEvent/blob/87422d850a3f0a4631528f5fbee23904170c0703/SystemEvent/__init__.py#L47
https://github.com/GEANT/CAT/blob/b53bc299c7e822c7abd8deb1ee1a9e44f3f465da/ansible/ManagedSP/templates/daemon/fr_restart.py#L46
https://github.com/south-coast-science/scs_host_rpi/blob/50b6277b4281b043d7c8f340371183faea4c5b8c/src/scs_host/sync/binary_semaphore.py#L54
                """
                while available_resources > self._sem.value:
                    self._sem.release()
                while available_resources < self._sem.value:
                    self._sem.acquire()

        finally:
            os.umask(old_umask)

    def acquire(self, timeout=DEFAULT_TIMEOUT):
        """Acquire the lock
        """
        try:
            self._sem.acquire(timeout)
        except posix_ipc.BusyError:
            raise ResourceBusyError()
        except posix_ipc.Error:  # Catch all IPC Errors except BusyError
            raise CannotLockError()

    def release(self):
        """Release the lock
        """
        self._sem.release()

    def locked(self):
        return self.value() == 0

    def value(self):
        return self._sem.value


class ResourceBusyError(posix_ipc.Error):
    """
    Raised when a call times out
    """

    def __init__(self, *args, **kwargs):
        pass


class CannotLockError(posix_ipc.Error):
    """
    Raised when can not lock the resource due to the permission issue,
    wrong IPC object or whatever internal issue.
    """

    def __init__(self, *args, **kwargs):
        pass


def ecc_lock(timeout=DEFAULT_TIMEOUT, raise_exception=False):
    """
    Returns an ECC LOCK decorator.

    timeout: timeout value. DEFAULT_TIMEOUT = 2 seconds.
    raise_exception: set True to raise exception in case of timeout and error.
                    Otherwise just log the error msg
    """

    def decorator_ecc_lock(func):
        lock = InterprocessLock(LOCK_ECC)

        @functools.wraps(func)
        def wrapper_ecc_lock(*args, **kwargs):
            try:
                # try to acquire the ECC resource or may raise an exception
                lock.acquire(timeout=timeout)

                value = func(*args, **kwargs)

                # release the resource
                lock.release()

                return value
            except ResourceBusyError as ex:
                LOGGER.error("ECC is busy now.")
                if raise_exception:
                    raise ex
            except CannotLockError as ex:
                LOGGER.error("Can't lock the ECC for some internal issue.")
                if raise_exception:
                    raise ex

        return wrapper_ecc_lock

    return decorator_ecc_lock



import unittest
import threading
import pytest
import mock
from time import sleep
from hm_pyhelper.interprocess_lock import ecc_lock, InterprocessLock, \
    ResourceBusyError, LOCK_ECC


# https://gist.github.com/sbrugman/59b3535ebcd5aa0e2598293cfa58b6ab
@pytest.fixture(autouse=True, scope="function")
def error_on_raise_in_thread():
    """
    Replaces Thread with a a wrapper to record any exceptions and
    re-raise them after test execution. In case multiple threads
    raise exceptions only one will be raised.
    """
    last_exception = None

    class ThreadWrapper(threading.Thread):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def run(self):
            try:
                super().run()
            except BaseException as e:
                nonlocal last_exception
                last_exception = e

    with mock.patch('threading.Thread', ThreadWrapper):
        yield
        if last_exception:
            raise last_exception


class TestInterprocessLock(unittest.TestCase):
    def test_interprocess_lock_basic(self):
        lock = InterprocessLock('test', available_resources=1, reset=True)

        self.assertFalse(lock.locked())
        lock.acquire()
        self.assertTrue(lock.locked())

        # Do some work
        sleep(0.001)

        lock.release()
        self.assertFalse(lock.locked())

    def test_interprocess_lock_timeout(self):
        """
        Start a slow running thread and then try to acquire a shared
        lock. Expect the acquire call to throw an exception.
        """
        lock = InterprocessLock('test', available_resources=1, reset=True)

        def slow_task():
            sleep(0.01)
            # lock.release()
            return True

        slow_thread = threading.Thread(target=slow_task, daemon=True)

        lock.acquire()
        slow_thread.start()

        expected_exception = False
        try:
            lock.acquire(timeout=0.00001)
        except ResourceBusyError:
            expected_exception = True

        self.assertTrue(expected_exception)

    def test_interprocess_lock_racing(self):
        def slow_task():
            lock = InterprocessLock('racing')

            lock.acquire()
            self.assertTrue(lock.locked())

            # Do some slow work
            print("Starting the slow task...")
            sleep(0.003)
            print("Finished the slow task!")

            lock.release()
            self.assertFalse(lock.locked())

        def fast_task():
            lock = InterprocessLock('racing')

            lock.acquire()
            self.assertTrue(lock.locked())

            # Do some slow work
            print("Starting the fast task...")
            sleep(0.001)
            print("Finished the fast task!")

            lock.release()
            self.assertFalse(lock.locked())

        lock = InterprocessLock('racing', available_resources=1, reset=True)

        slow_thread = threading.Thread(target=slow_task, daemon=True)
        fast_thread = threading.Thread(target=fast_task, daemon=True)

        # Ensure there is no lock before starting the work
        self.assertFalse(lock.locked())

        # Start work
        print("\n")
        print("Initiated the slow task.")
        slow_thread.start()
        print("Initiated the fast task.")
        fast_thread.start()

        # Wait to finish
        fast_thread.join()
        slow_thread.join()

        # Ensure there is no lock after finishing the work
        self.assertFalse(lock.locked())

    def test_ecc_lock_basic(self):
        @ecc_lock()
        def some_task():
            sleep(0.00001)

        lock = InterprocessLock(LOCK_ECC, available_resources=1, reset=True)
        self.assertFalse(lock.locked())

        some_task_thread = threading.Thread(target=some_task, daemon=True)
        some_task_thread.start()
        self.assertTrue(lock.locked())

        some_task_thread.join()
        self.assertFalse(lock.locked())

    def test_ecc_lock_timeout(self):
        @ecc_lock()
        def slow_task():
            sleep(0.1)

        @ecc_lock(timeout=0.01, raise_exception=True)
        def lock_with_timeout():
            pass

        lock = InterprocessLock(LOCK_ECC, available_resources=1, reset=True)
        self.assertFalse(lock.locked())

        slow_thread = threading.Thread(target=slow_task, daemon=True)
        slow_thread.start()

        expected_exception = False
        try:
            lock_with_timeout()
        except ResourceBusyError:
            expected_exception = True

        self.assertTrue(expected_exception)

    def test_ecc_lock_racing(self):
        @ecc_lock()
        def slow_task():
            # Do some slow work
            print("Starting the slow task...")
            sleep(0.003)
            print("Finished the slow task!")

        @ecc_lock()
        def fast_task():
            # Do some slow work
            print("Starting the fast task...")
            sleep(0.001)
            print("Finished the fast task!")

        lock = InterprocessLock(LOCK_ECC, available_resources=1, reset=True)
        self.assertFalse(lock.locked())

        slow_thread = threading.Thread(target=slow_task, daemon=True)
        fast_thread = threading.Thread(target=fast_task, daemon=True)

        # Start work
        print("\n")
        print("Initiated the slow task.")
        slow_thread.start()
        print("Initiated the fast task.")
        fast_thread.start()

        # Wait to finish
        fast_thread.join()
        slow_thread.join()

        # Ensure there is no lock after finishing the work
        self.assertFalse(lock.locked())
